/* 
 * This is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
 * USA.
 */
package android.androidVNC;

import java.util.ArrayList;
import java.util.List;
import java.util.concurrent.Semaphore;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import android.app.AlertDialog;
import android.app.ListActivity;
import android.content.Context;
import android.content.DialogInterface;
import android.content.Intent;
import android.os.Bundle;
import android.util.Log;
import android.view.ContextMenu;
import android.view.Menu;
import android.view.MenuItem;
import android.view.View;
import android.view.ContextMenu.ContextMenuInfo;
import android.view.View.OnClickListener;
import android.widget.AdapterView;
import android.widget.ArrayAdapter;
import android.widget.Button;
import android.widget.EditText;
import android.widget.ListView;
import android.widget.Toast;
import android.widget.AdapterView.OnItemLongClickListener;

/**
 * <p>(Se omiten acentos)</p>
 * <p>Clase que muestra por pantalla el arbol de ficheros del servidor remoto permitiendo elegir la ubicacion donde se
 *    copiara el archivo local. Tambien permite crear carpetas y eliminar archivos remotos.</p>
 * <p>Fichero: ExploradorArchivosRemoto.java.</p>
 * <p>Aadido al PFC en la version v1.0.</p>
 * <p><b>Proyecto original:</b></p>
 * 		<p>Web: http://code.google.com/p/android-vnc-viewer/.</p>
 * 		<p>Release: 39.</p>
 *	 	<p>Migrada a release 60.</p>
 * @author Alberto Poblacion Navarro.
 */
public class ExploradorArchivosRemoto extends ListActivity{
	/** Cadena para identifcar la clase en el log interno*/
	private static final String TAG = "ExploradorArchivosRemoto";
	/** Cadena para identificar el tipo de informacion en el intent*/
	public static final String TIPO = "tipo";
	/** Tipo de funcionamiento de la clase: Enviar archivo al servidor */
	public static final int ENVIAR = 1; 
	/** Tipo de funcionamiento de la clase: Recibir archivo del servidor */
	public static final int RECIBIR = 2;
	/** Lista de cadenas con los nombre de los ficheros de un directorio */
	private List<String> elementos = null;
	/** Variable que almacena el tipo de funcionamiento de la clase */
	private int tipo; 
	/** Objeto utilizado para enviar y recibir archivos */
    private FileTransfer ft;
    /** Cadena utilizada para almacenar la direccion del archivo */
    String path= new String(); 
	/** Array de cadenas donde se almacena el arbol de archivos de un directorio*/
    static String[] listado; 
    /** Semaforo utilizado para la sincronizacion de esta clase con el objeto de FileTransfer */
	static Semaphore semaforo = new Semaphore(0);
	/** Variable para conocer a partir de que posicion del array se listan los ficheros y no directorios */
	public static int primer_archivo=1000;
	/** Boton que se mostrara en primer lugar, su funcionalidad depende del proposito de la clase (tipo) */
	private Button boton_uno;
	/** Boton que se mostrara en segundo lugar, su funcionalidad depende del proposito de la clase (tipo) */
	private Button boton_dos;
	/** Posicion de la ListView seleccionada */
	private int posicion;
	/** Vista utilizada para listar los archivos */
	private ListView listView;
	/** Cadena donde se almacena el nombre del archivo seleccionado para una comparacion con el siguiente que seleccionemos */
	String anterior=null;
	/** Contexto de ejecucion*/
	Context context = this;
	
	/**
     * Funcion llamada automaticamente cuando se lanza la actividad.
     * Se muestra la raiz de ficheros del ordenador local: Mi PC.
     * (Esta caracteristica solo funciona con servidores UltraVNC, por lo tanto solo esta disponible en Windows)
     * Segun el funcionamiento elegido de la clase se mostrara la lista de ficheros para elegir archivo o directorio.
     * Se configuran los botones de acuerdo al tipo de funcionamiento elegido. 
     */
	public void onCreate(Bundle icicle) {
	        super.onCreate(icicle);
	        semaforo.drainPermits();
	        semaforo = new Semaphore(0);
	        //Se da un valor alto al primer_archivo para que no se produzcan errores. El valor correcto se da cuando se rellena la lista
	        primer_archivo=10000;
	        //Mostramos la lista con los archivos que se obtienen en rellanar
	        setContentView(R.layout.listado);
	        //Obtenemos la instancia que vamos a utilizar de FileTransfer y el tipo de explorador remoto que necesitamos
	        //Utilizamos la misma instancia que la utilizada en la clase VncCanvasActivity.
	        ft=VncCanvasActivity.getFileTransfer();
	        tipo=getIntent().getIntExtra(TIPO, 2);
	        //Si se va a realizar un envio al servidor
	    	if(tipo == ENVIAR){
	    		//Se muestra una cadena con la informacin
	    		Toast.makeText(this, getString(R.string.file_remote), Toast.LENGTH_LONG).show();
	    		//Se configuran los botones.
				boton_uno = (Button)findViewById(R.id.boton1);
				boton_uno.setText(getString(R.string.copy_here));
				boton_uno.setEnabled(false);
				boton_uno.setOnClickListener(new OnClickListener(){
				
					@Override
					public void onClick(View v) {
						//Se crea un intent con la informacion que se pasara a la actividad que recoja el retorno.
						Intent intent = new Intent();
						intent.putExtra("tipo", tipo);
			       	 	intent.putExtra("path", path);
			       	 	setResult(RESULT_OK, intent);
			       	 	finish();
					}
				});
				boton_dos= (Button)findViewById(R.id.boton2);
				boton_dos.setText(getString(android.R.string.cancel));
				boton_dos.setOnClickListener(new OnClickListener(){
					@Override
					public void onClick(View v) {
						setResult(RESULT_CANCELED);
						finish();
					}
				});
	    		
			}
	    	if(tipo==RECIBIR){ //Si se va a copiar un archivo del servidor.
	    		setTitle("Seleccione el archivo a copiar");
	    		
	    		//Configuracion de los botones
				boton_uno = (Button)findViewById(R.id.boton1);
				boton_uno.setText("Enviar");
				boton_uno.setEnabled(false);
				boton_uno.setOnClickListener(new OnClickListener(){
					@Override
					public void onClick(View v) {
						Intent intent = new Intent();
	        			intent.putExtra("tipo", tipo);
			       	 	intent.putExtra("path", path+"\\"+anterior);
			       	 	intent.putExtra("archivo", anterior);
			       	 	setResult(RESULT_OK, intent);
			       	  	finish();
						}
					
				});
				boton_dos= (Button)findViewById(R.id.boton2);
				boton_dos.setText(getString(android.R.string.cancel));
				boton_dos.setOnClickListener(new OnClickListener(){
					@Override
					public void onClick(View v) {
						setResult(RESULT_CANCELED);
						finish();
					}
				});
				
	    	}
	    	listView = getListView();
	        listView.setChoiceMode(ListView.CHOICE_MODE_SINGLE);
			//Si se produce una pulsacion larga en un elemento de la lista se mostrara un menu con opciones
			listView.setOnItemLongClickListener(new OnItemLongClickListener(){
				@Override
				public boolean onItemLongClick(AdapterView<?> arg0, View arg1,
						int arg2, long arg3) {
					posicion=arg2;
					registerForContextMenu(arg0);
					return false;
				}
	        	
	        });
	    	rellenarConElRaiz();
	}
	/**
	 * Se crea un menu con los elementos descritos en el archivo de recursos.
	 * Ese menu solo estara disponible en caso si el fin de la clase es enviar un archivo a la clase.
	 * @param menu Menu creado
	 */
	@Override
	public boolean onCreateOptionsMenu(Menu menu) {
		if(tipo==ENVIAR)
			getMenuInflater().inflate(R.menu.exploradormenu, menu);
		return true;
	}
	/**
	 * Si se selecciona un una opcion del menu se lanzara automaticamente esta funcion.
	 * @param item Opcion seleccionada
	 */
	@Override
	public boolean onOptionsItemSelected(MenuItem item) {
		switch (item.getItemId())
		{
		//Mostramos la opcion de crear una carpeta
		case R.id.itemCreateDirectory:
			//Cuadro de edicion donde se escribira el nombre de la nueva carpeta
			final EditText edittext = new EditText(this);
			edittext.setText(null);
			//Cuadro de dialogo mostrado para crear un nuevo nombre a la carpeta.
			new AlertDialog.Builder(this)
            .setTitle(getString(R.string.new_directory)+path)
            .setMessage(getString(R.string.new_name))
            .setView(edittext)
            //Si se acepta se crea una nueva carpeta remota con el nombre indicado y se actualiza la vista
            .setPositiveButton(android.R.string.ok, new DialogInterface.OnClickListener(){
                public void onClick(DialogInterface dialog, int whichButton) {
                	if(edittext.getText().toString() != null){
                		//Se crea un nuevo directorio con el nombre que se ha introducido en el cuadro de edicion
                		ft.createRemoteDirectory(path+"\\"+edittext.getText());
                		actualizar_vista(path);
                	}
               	 
               }})
            .setNegativeButton(android.R.string.cancel, null)
            .create().show();
			break;
		}
		return true;
	}
	
	/**
	 * Funcion cuando se finalice la activiy.
	 * Si no se ha elegido ningun archivo se establecer el retorno como tal.
	 * Se eliminan los permisos que pueden existir en el semaforo para futuras ejecuciones.
	 * @author Alberto Poblacion Navarro
	 */
    public void onDestroy() {
    	super.onDestroy();
    	Log.d(TAG, "OnDestroy");
    	setResult(RESULT_CANCELED, null);
    	path=null;
    	listado = null;
    	semaforo.drainPermits();
    	
    }
     /**
	 * Muestra el listado de ficheros desde la raiz
	 * @author Alberto Poblacion Navarro
	 */
	private void rellenarConElRaiz() {
				setTitle("MI PC");
				ft.readServerDriveList();
				primer_archivo=1000;
				path=new String();
				try {
					//Esperamos a que recibamos la lista de carpetas y se libere el semaforo
					semaforo.acquire();
				} catch (InterruptedException e) {
					Log.d(TAG, "EXCEPCION: "+e.toString());
					path=null;
			    	listado = null;
			    	semaforo.drainPermits();
					
				}
		        rellenar(listado);
		        boton_uno.setEnabled(false);
	}
	/**
	 * Se invoca automaticamente esta funcion cuando se seleccione una opcion de la lista.
	 * @param l Vista de la lista
	 * @param v Vista asiciada a la vista
	 * @param position Posicion seleccionada
	 * @para id Identificador del objeto seleccionado
	 */
	@Override
    protected void onListItemClick(ListView l, View v, final int position, long id) {
        int IDFilaSeleccionada = position;
        if (IDFilaSeleccionada==0){
            rellenarConElRaiz();
        } else {
        		Pattern p = Pattern.compile("[a-zA-Z0-9._\\s\\p{Punct}-()&&[^]]*");
				Matcher m = p.matcher(l.getItemAtPosition(position).toString());
				if(m.matches()){
	        		boton_uno.setEnabled(false);
	        		//Si el archivo es una carpeta mostramos su contenido
	        		if(IDFilaSeleccionada < primer_archivo){
		        			if(tipo == ENVIAR) boton_uno.setEnabled(true);
				           	path=path+elementos.get(IDFilaSeleccionada)+"\\";
				           	setTitle(path);
				        	ft.readServerDirectory(path);
				        	try {
				        		//Se para el programa hasta que se reciba la lista de carpetas y se libere el semaforo
								semaforo.acquire(); 
							} catch (InterruptedException e) {
								e.printStackTrace();
								path=null;
						    	listado = null;
						    	semaforo.drainPermits();
							}
							//Si cuando libere el semaforo la variable se ha puesto a false, quiere decir que el 
							//directorio deseado no se encuentra disponible. Mostramos de nuevo la raiz
							if(!ft.inDirectory2){
								Toast.makeText(this, getString(R.string.no_directory), Toast.LENGTH_SHORT).show();
								rellenarConElRaiz();
							}else //Si todo va bien, mostramos el directorio elegido
								rellenar(listado);
		        	}else{
						 if(tipo == RECIBIR){
							 //Si no se ha seleccionado anteriormente no entramos nuevamente
							if(l.getItemAtPosition(position).toString() != anterior){
								boton_uno.setEnabled(true);
								anterior = l.getItemAtPosition(position).toString().substring(1);
								posicion = position;
				        		listView.setItemChecked(position, true);
				        		l.setItemChecked(position, true);
							}	
		        		}
		        	}
				}else{
					Toast.makeText(context, getString(R.string.codification_problem), Toast.LENGTH_SHORT).show();
					boton_uno.setEnabled(false);
				}
        }
    }
	/**
	 * Actualiza la lista con la direccion indicada por parametro
	 * @param path Direccion remota para rellenar la vista
	 */
	void actualizar_vista(String path){
		ft.readServerDirectory(path);
    	try {
    		//Se para el programa hasta que se reciba la lista de carpetas y se libere el semaforo
			semaforo.acquire(); 
		} catch (InterruptedException e) {
			e.printStackTrace();
			path=null;
	    	listado = null;
	    	semaforo.drainPermits();
		}
		rellenar(listado);
	}
	/**
	 * Muestra por pantalla el la lista de ficheros contenida en el array de cadenas
	 * @param archivos Vector de cadenas con el nombre de los ficheros.
	 */
	void rellenar(String[] archivos) {
		elementos = new ArrayList<String>();
		elementos.add("[MI PC]");
        for(String archivo: archivos){
        	Log.d(TAG, "Archivo: "+archivo);
            elementos.add(archivo);
        }
        ArrayAdapter<String> listaArchivos;
        //Mostramos una lista de los archivo dependiendo del tipo de ejecucion.
        if(tipo == RECIBIR){
        	listaArchivos= new ArrayAdapter<String>(this, android.R.layout.simple_list_item_single_choice, elementos);
        }else{
        	listaArchivos= new ArrayAdapter<String>(this, android.R.layout.simple_list_item_1, elementos);        
        }
        setListAdapter(listaArchivos);
    }
	/**
	 * Funcion que se llama cuando se lanza un menu conextual.
	 * Se llamara cuando se deje pulsada una opcion de la ListView durante un tiempo prolongado. 
	 * Dependiendo de si selecionamos una carpeta o un archivo mostrara una opcion u otra.
	 */
    @Override
    public void onCreateContextMenu(ContextMenu menu, View v,
    		ContextMenuInfo menuInfo) {
    	super.onCreateContextMenu(menu, v, menuInfo);
    	menu.clear();
    	menu.setHeaderTitle("Opciones");
    	if(posicion < primer_archivo)
    		menu.add(0,2,Menu.NONE,"Abrir");
    	else{ 
    		menu.add(0,1,Menu.NONE,"Seleccionar");
    		menu.add(0,0,Menu.NONE,"Eliminar");
    	}
    	
    }
    /**
     * Funcion llamada automaticamente cuando se seleccione una opcion del menu contextual.
     * Permite eliminar un archivo, seleccionarlo o abrir un directorio.
     * @param item Opcion seleccionada
     */
    @Override
    public boolean onContextItemSelected(MenuItem item) {
    	//Se omiten caracteres raros por incopatibilidad de codificacion entre Android y Windows
    	Pattern p = Pattern.compile("[a-zA-Z0-9._\\s\\p{Punct}-()&&[^]]*");
		Matcher m = p.matcher(listView.getItemAtPosition(posicion).toString());
		if(m.matches()){
	    	switch (item.getItemId()) {
			case 0:
				ft.deleteRemoteFile(path+"\\"+elementos.get(posicion).substring(1, elementos.get(posicion).length()));
	         	actualizar_vista(path);
				break;
			case 1:
				listView.setItemChecked(posicion, true);
				boton_uno.setEnabled(true);
				break;
			case 2:
				path=path+elementos.get(posicion)+"\\";
	           	setTitle(path);
	        	ft.readServerDirectory(path);
	        	try {
	        		//Se para el programa hasta que se reciba la lista de carpetas y se libere el semaforo
					semaforo.acquire(); 
				} catch (InterruptedException e) {
					e.printStackTrace();
					path=null;
			    	listado = null;
			    	semaforo.drainPermits();
				}
				//Si cuando libere el semaforo la variable se ha puesto a false, quiere decir que el 
				//directorio deseado no se encuentra disponible. Se muestra de nuevo la raiz
				if(!ft.inDirectory2){
					Toast.makeText(this, getString(R.string.no_directory), Toast.LENGTH_SHORT).show();
					rellenarConElRaiz();
				}else //Si todo va bien, se muestra el directorio elegido
					rellenar(listado);
				break;
	    
	    	}
		}else Toast.makeText(this, getString(R.string.codification_problem), Toast.LENGTH_SHORT).show();
    	return super.onContextItemSelected(item);
    }
	
	
}

	